==> privacy-sensitive-analytics-v0.2.tar.bz2: privacy-sensitive-analytics-v0.2/server/analytics.lua <==
local table = table
local math = math
local require = require
local string = string
local http = require("resty.http")
local json = require "cjson"

function strSplit(delim,str)
    local t = {}

    for substr in string.gmatch(str, "[^".. delim.. "]*") do
        if substr ~= nil and string.len(substr) > 0 then
            table.insert(t,substr)
        end
    end

    return t
end

function table_contains(tbl, x)
    found = false
	for _, v in pairs(tbl) do
		if v == x then 
            found = true 
        end
	end
	return found
end

function tableHasKey(table,key)
    return table[key] ~= nil
end


function write_lp(server, port, lp, db)
    -- Initiate the HTTP connector
    local httpc = http.new()
    httpc:connect(server, port)

    -- We want to time out pretty quickly to avoid holding the request up too much
    httpc:set_timeouts(100, 100, 400)

    local res, err = httpc:request {
        path = '/write?precision=ms&db='..db,
        method = 'POST',
        body = lp
    }

    -- Return the connection to the KA pool
    httpc:set_keepalive()

    if not res then
        return False
    end
    
    -- Check the status, dispatcher will serve us a 302
    if res.status == 200
    then
        return True
    end
    
    return False
end


function build_unauth_lp(post, ts, measurement)
    -- Build LP to record a request relating to an unauth'd domain
    -- websites/privacy-sensitive-analytics#3
    
    sec1 = {
        measurement .. "_unauth",
        'type="unauthorised_domain"',
        'page="' .. post['page'] .. '"'
        }
    
    fields = {       
        'domain="' .. string.lower(post['domain']) .. '"',
        "timezone=" .. post['timezone'] .. "i",
        'referrer="' .. post['referrer'] .. '"',
        'referrer_domain="' .. string.lower(post['referrer_domain']) .. '"',
        'platform="' .. string.lower(post['platform']) .. '"',
        "response_time=" .. post['responseTime'] .. "i"       
        }
    
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    return table.concat(join_t, " ") 
end


function build_lp(ngx, measurement)
    local get, post, files = require "resty.reqargs"()
    -- post is a table containing the decoded JSON

    -- Run safety checks
    validate_write(ngx, post)
    
    -- Skip domains specifies domains that we do nothing for - just return a 401
    skip_domains = strSplit(",", ngx.var.skip_domains)
    if table_contains(skip_domains, post['domain']) then
       ngx.exit(401) 
    end
    
    local ts = ngx.now() * 1000
    
    allowed_domains = strSplit(",", ngx.var.permitted_domains)
    
    if not table_contains(allowed_domains, string.lower(post['domain'])) then
        -- Content being served from somewhere unauthorised
        -- websites/privacy-sensitive-analytics#3
        return build_unauth_lp(post, ts, measurement), false
    end
    
    -- TODO: do we want to include the first bit of the path as a tag?
    -- perhaps the first 2 directories deep?
    sec1 = {
        measurement,
        "domain=" .. string.lower(post['domain']),
        'type="pageview"',
        'page="' .. post['page'] .. '"'
        }
    
    if tableHasKey(post, "sess_id") and post["sess_id"] ~= "none" then
       table.insert(sec1, 'sess_id="' .. post["sess_id"] .. '"')
    end
    
    if tableHasKey(post, "state") then
       table.insert(sec1, 'action="' .. post["state"] .. '"')
    end   
    
    fields = {       
        "timezone=" .. post['timezone'] .. "i",
        'referrer="' .. post['referrer'] .. '"',
        'referrer_domain="' .. string.lower(post['referrer_domain']) .. '"',
        'platform="' .. string.lower(post['platform']) .. '"',
        "response_time=" .. post['responseTime'] .. "i"
        }

    -- Put them together
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    
    -- Final concat
    return table.concat(join_t, " "), true
end

function build_searchterm_lp(ngx, measurement)
    local get, post, files = require "resty.reqargs"()
    -- post is a table containing the decoded JSON

    -- Run safety checks
    -- TODO: create search terms specific version
    -- validate_write(ngx, post)
    
    -- Skip domains specifies domains that we do nothing for - just return a 401
    skip_domains = strSplit(",", ngx.var.skip_domains)
    if table_contains(skip_domains, post['domain']) then
       ngx.exit(401)
    end
    
    local ts = ngx.now() * 1000
    
    allowed_domains = strSplit(",", ngx.var.permitted_domains)
    
    if not table_contains(allowed_domains, string.lower(post['domain'])) then
        -- Refuse the write
        ngx.exit(403)
    end
    
    -- Set up measurement and tags
    sec1 = {
        measurement,
        "domain=" .. string.lower(post['domain']),
        'type=search'
        }
    
    if tableHasKey(post, "sess_id") and post["sess_id"] ~= "none" then
       table.insert(sec1, 'sess_id="' .. post["sess_id"] .. '"')
    end
    
    fields = {
        'search_term="' .. post['search_term'] .. '"'
        }

    if tableHasKey(post, "result_count") then
       table.insert(fields, 'result_count=' .. post["result_count"] .. 'i')
    end
    
    -- Put them together
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    
    -- Final concat
    return table.concat(join_t, " ")
end

function build_enhanced_lp(ngx, measurement)
    local get, post, files = require "resty.reqargs"()
    
    -- Skip domains specifies domains that we do nothing for - just return a 401
    -- TODO: I'd like to be able to do this by suffix
    skip_domains = strSplit(",", ngx.var.skip_domains)
    if table_contains(skip_domains, post['domain']) then
       ngx.exit(401)
    end
    
    local ts = ngx.now() * 1000
    
    -- No authorised domains check here, this write is triggered when
    -- an unauthorised domain is detected
    -- Set up measurement and tags
    sec1 = {
        measurement .. "_unauth",
        'type="enhancedwrite"',
        'page="' .. post['page'] .. '"'
        }
    
    
    -- We don't want to post the screenshot upstream, that way lays sadness
    fname = "/mnt/images/screenshot_" .. tostring(ts) .. ".png"
    imgfile = io.open(fname , "w")
    l = string.gsub(post['screenshot'], "data:image/png;base64,", "")
    ngx.log(ngx.ERR,l)
    s = ngx.decode_base64(l)
    imgfile:write(s)
    imgfile:close()
    
    fields = {       
        'domain="' .. string.lower(post['domain']) .. '"',
        "timezone=" .. post['timezone'] .. "i",
        'referrer="' .. post['referrer'] .. '"',
        'platform="' .. string.lower(post['platform']) .. '"',
        "response_time=" .. post['responseTime'] .. "i" ,
        'page_title="' .. post['pageTitle'] .. '"',
        'screenshot="' .. fname .. '"'
        }
    
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    return table.concat(join_t, " ")  
end

function validate_write(ngx, post)
    -- Take some precautions to try and keep junk out of the DB
    
    if string.sub(post['page'], 1, 1) ~= "/" then
        -- Invalid URL
        ngx.header['x-reason'] = 'page'
        ngx.exit(400)
    end
    
    -- Lengths
    if string.len(post['page']) > 300 or string.len(post['domain']) > 120 or string.len(post['referrer_domain']) > 120 or string.len(post['referrer']) > 300 then
       ngx.header['x-reason'] = 'badlen'
       ngx.exit(400)
    end
    
    -- Page paths
    if string.find(post['page'], " ") or string.find(post['page'], ";") or string.find(post['page'], "UNION") then
        ngx.header['x-reason'] = 'naughtypage'
        ngx.exit(400)                                                     
    end

    -- Domains
    if string.find(post['domain'], " ") or string.find(post['domain'], ";") or string.find(post['domain'], "UNION") or string.find(post['domain'], "/") then
        ngx.header['x-reason'] = 'naughtydomain'
        ngx.exit(400)                                                     
    end
    if string.find(post['referrer_domain'], " ") or string.find(post['referrer_domain'], ";") or string.find(post['referrer_domain'], "UNION") or string.find(post['referrer_domain'], "/") then
        ngx.header['x-reason'] = 'naughtyreferrer'
        ngx.exit(400)                                                     
    end    

    -- URLs
    if post['referrer'] ~= "" and not string.match(post['referrer'], '[a-z]*://[^ >,;]*') then
       ngx.header['x-reason'] = 'nourl'
       ngx.exit(400)
    end
    
    -- Session ID
    if tableHasKey(post, "sess_id") and post["sess_id"] ~= "none" and not string.match(post["sess_id"], '[%a%d][%a%d][%a%d][%a%d]%-[%a%d][%a%d][%a%d][%a%d]') then
        ngx.header['x-reason'] = "sess"
        ngx.exit(400)
    end    
    
    -- State
    local states = {"PageView", "ready", "playing"}
    if tableHasKey(post, "state") and not table_contains(states, post["state"]) then
        ngx.header['x-reason'] = "state"
        ngx.exit(400)
    end        
    
    
    -- Numbers
    if not tonumber(post['timezone']) then
       ngx.header['x-reason'] = 'tz'
       ngx.exit(400)
    end
    if post['timezone'] > 1440 or post['timezone'] < -1440 then
       ngx.header['x-reason'] = 'tzrange'
       ngx.exit(400) 
    end
    
    if not tonumber(post['responseTime']) then
       ngx.header['x-reason'] = 'resp'
       ngx.exit(400)
    end    
    if post['responseTime'] > 900000 or post['responseTime'] < 0 then
       ngx.header['x-reason'] = 'resprange'
       ngx.exit(400) 
    end
end


local server = ngx.var.influx_ip
local port = ngx.var.influx_port
local measurement = ngx.var.measurement
local db = ngx.var.db
local mode = ngx.var.mode

if mode == "write_search" then
    local lp = build_searchterm_lp(ngx, measurement)
    write_lp(server, port, lp, db)
    ngx.say(lp)
else
    if mode == "write_enhanced" then
        local lp = build_enhanced_lp(ngx, measurement)
        write_lp(server, port, lp, db)
        ngx.say('')        
    else
        local lp, result = build_lp(ngx, measurement)
        write_lp(server, port, lp, db)
        
        if not result then
           ngx.status = 203
           ngx.header['Access-Control-Allow-Origin'] = '*'
           ngx.header['Access-Control-Allow-Methods'] = 'POST, OPTIONS'
           ngx.say("unoh")
           ngx.exit(203)
        end
        
        ngx.say(lp)
    end
end

==> privacy-sensitive-analytics-v0.4.tar.bz2: privacy-sensitive-analytics-v0.4/server/analytics.lua <==
local table = table
local math = math
local require = require
local string = string
local http = require("resty.http")
local json = require "cjson"

function strSplit(delim,str)
    local t = {}

    for substr in string.gmatch(str, "[^".. delim.. "]*") do
        if substr ~= nil and string.len(substr) > 0 then
            table.insert(t,substr)
        end
    end

    return t
end

function table_contains(tbl, x)
    found = false
	for _, v in pairs(tbl) do
		if v == x then 
            found = true 
        end
	end
	return found
end

function tableHasKey(table,key)
    return table[key] ~= nil
end


function write_lp(server, port, lp, db)
    -- Initiate the HTTP connector
    local httpc = http.new()
    httpc:connect(server, port)

    -- We want to time out pretty quickly to avoid holding the request up too much
    httpc:set_timeouts(100, 100, 400)

    local res, err = httpc:request {
        path = '/write?precision=ms&db='..db,
        method = 'POST',
        body = lp
    }

    -- Return the connection to the KA pool
    httpc:set_keepalive()

    if not res then
        return False
    end
    
    -- Check the status, dispatcher will serve us a 302
    if res.status == 200
    then
        return True
    end
    
    return False
end


function build_unauth_lp(post, ts, measurement)
    -- Build LP to record a request relating to an unauth'd domain
    -- websites/privacy-sensitive-analytics#3
    
    sec1 = {
        measurement .. "_unauth",
        'type="unauthorised_domain"',
        'page="' .. post['page'] .. '"'
        }
    
    fields = {       
        'domain="' .. string.lower(post['domain']) .. '"',
        "timezone=" .. post['timezone'] .. "i",
        'referrer="' .. post['referrer'] .. '"',
        'referrer_domain="' .. string.lower(post['referrer_domain']) .. '"',
        'platform="' .. string.lower(post['platform']) .. '"',
        "response_time=" .. post['responseTime'] .. "i"       
        }
    
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    return table.concat(join_t, " ") 
end


function build_404_lp(ngx, measurement)
    local get, post, files = require "resty.reqargs"()
    -- post is a table containing the decoded JSON

    -- Run safety checks
    validate_write(ngx, post)
    
    -- Skip domains specifies domains that we do nothing for - just return a 401
    skip_domains = strSplit(",", ngx.var.skip_domains)
    if table_contains(skip_domains, post['domain']) then
       ngx.exit(401) 
    end
    
    local ts = ngx.now() * 1000
    
    allowed_domains = strSplit(",", ngx.var.permitted_domains)
    
    if not table_contains(allowed_domains, string.lower(post['domain'])) then
        -- Content being served from somewhere unauthorised
        -- websites/privacy-sensitive-analytics#3
        return build_unauth_lp(post, ts, measurement), false
    end
    
    -- TODO: do we want to include the first bit of the path as a tag?
    -- perhaps the first 2 directories deep?
    sec1 = {
        measurement,
        "domain=" .. string.lower(post['domain']),
        'resp_status=404',
        'page="' .. post['page'] .. '"'
        }
    
    if tableHasKey(post, "sess_id") and post["sess_id"] ~= "none" then
       table.insert(sec1, 'sess_id="' .. post["sess_id"] .. '"')
    end
    
    if tableHasKey(post, "state") then
       table.insert(sec1, 'action="' .. post["state"] .. '"')
    end   
    
    fields = {       
        "timezone=" .. post['timezone'] .. "i",
        'referrer="' .. post['referrer'] .. '"',
        'referrer_domain="' .. string.lower(post['referrer_domain']) .. '"',
        'platform="' .. string.lower(post['platform']) .. '"',
        "response_time=" .. post['responseTime'] .. "i"
        }

    -- Put them together
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    
    -- Final concat
    return table.concat(join_t, " "), true
    
end


function build_lp(ngx, measurement)
    local get, post, files = require "resty.reqargs"()
    -- post is a table containing the decoded JSON

    -- Run safety checks
    validate_write(ngx, post)
    
    -- Skip domains specifies domains that we do nothing for - just return a 401
    skip_domains = strSplit(",", ngx.var.skip_domains)
    if table_contains(skip_domains, post['domain']) then
       ngx.exit(401) 
    end
    
    local ts = ngx.now() * 1000
    
    allowed_domains = strSplit(",", ngx.var.permitted_domains)
    
    if not table_contains(allowed_domains, string.lower(post['domain'])) then
        -- Content being served from somewhere unauthorised
        -- websites/privacy-sensitive-analytics#3
        return build_unauth_lp(post, ts, measurement), false
    end
    
    -- TODO: do we want to include the first bit of the path as a tag?
    -- perhaps the first 2 directories deep?
    sec1 = {
        measurement,
        "domain=" .. string.lower(post['domain']),
        'type="pageview"',
        'page="' .. post['page'] .. '"'
        }
    
    if tableHasKey(post, "sess_id") and post["sess_id"] ~= "none" then
       table.insert(sec1, 'sess_id="' .. post["sess_id"] .. '"')
    end
    
    if tableHasKey(post, "state") then
       table.insert(sec1, 'action="' .. post["state"] .. '"')
    end
    
    if tableHasKey(post, "protocol") then
       table.insert(sec1, 'protocol="' .. post["state"] .. '"')
    end   
    
    fields = {       
        "timezone=" .. post['timezone'] .. "i",
        'referrer="' .. post['referrer'] .. '"',
        'referrer_domain="' .. string.lower(post['referrer_domain']) .. '"',
        'platform="' .. string.lower(post['platform']) .. '"',
        "response_time=" .. post['responseTime'] .. "i"
        }

    -- Put them together
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    
    -- Final concat
    return table.concat(join_t, " "), true
end

function build_searchterm_lp(ngx, measurement)
    local get, post, files = require "resty.reqargs"()
    -- post is a table containing the decoded JSON

    -- Run safety checks
    -- TODO: create search terms specific version
    -- validate_write(ngx, post)
    
    -- Skip domains specifies domains that we do nothing for - just return a 401
    skip_domains = strSplit(",", ngx.var.skip_domains)
    if table_contains(skip_domains, post['domain']) then
       ngx.exit(401)
    end
    
    local ts = ngx.now() * 1000
    
    allowed_domains = strSplit(",", ngx.var.permitted_domains)
    
    if not table_contains(allowed_domains, string.lower(post['domain'])) then
        -- Refuse the write
        ngx.exit(403)
    end
    
    -- Set up measurement and tags
    sec1 = {
        measurement,
        "domain=" .. string.lower(post['domain']),
        'type=search'
        }
    
    if tableHasKey(post, "sess_id") and post["sess_id"] ~= "none" then
       table.insert(sec1, 'sess_id="' .. post["sess_id"] .. '"')
    end
    
    fields = {
        'search_term="' .. post['search_term'] .. '"'
        }

    if tableHasKey(post, "result_count") then
       table.insert(fields, 'result_count=' .. post["result_count"] .. 'i')
    end
    
    -- Put them together
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    
    -- Final concat
    return table.concat(join_t, " ")
end

function build_enhanced_lp(ngx, measurement)
    local get, post, files = require "resty.reqargs"()
    
    -- Skip domains specifies domains that we do nothing for - just return a 401
    -- TODO: I'd like to be able to do this by suffix
    skip_domains = strSplit(",", ngx.var.skip_domains)
    if table_contains(skip_domains, post['domain']) then
       ngx.exit(401)
    end
    
    local ts = ngx.now() * 1000
    
    -- No authorised domains check here, this write is triggered when
    -- an unauthorised domain is detected
    -- Set up measurement and tags
    sec1 = {
        measurement .. "_unauth",
        'type="enhancedwrite"',
        'page="' .. post['page'] .. '"'
        }
    
    if tableHasKey(post, "protocol") then
       table.insert(sec1, 'protocol="' .. post["state"] .. '"')
    end    
    
    fields = {       
        'domain="' .. string.lower(post['domain']) .. '"',
        "timezone=" .. post['timezone'] .. "i",
        'referrer="' .. post['referrer'] .. '"',
        'platform="' .. string.lower(post['platform']) .. '"',
        "response_time=" .. post['responseTime'] .. "i" ,
        'page_title="' .. post['pageTitle'] .. '"'
        }
    
    join_t = {
        table.concat(sec1, ","),
        table.concat(fields, ","),
        ts
        }
    return table.concat(join_t, " ")  
end

function validate_write(ngx, post)
    -- Take some precautions to try and keep junk out of the DB
    
    if string.sub(post['page'], 1, 1) ~= "/" then
        -- Invalid URL
        ngx.header['x-reason'] = 'page'
        ngx.exit(400)
    end
    
    -- Lengths
    if string.len(post['page']) > 300 or string.len(post['domain']) > 120 or string.len(post['referrer_domain']) > 120 or string.len(post['referrer']) > 300 then
       ngx.header['x-reason'] = 'badlen'
       ngx.exit(400)
    end
    
    -- Page paths
    if string.find(post['page'], " ") or string.find(post['page'], ";") or string.find(post['page'], "UNION") then
        ngx.header['x-reason'] = 'naughtypage'
        ngx.exit(400)                                                     
    end

    -- Domains
    if string.find(post['domain'], " ") or string.find(post['domain'], ";") or string.find(post['domain'], "UNION") or string.find(post['domain'], "/") then
        ngx.header['x-reason'] = 'naughtydomain'
        ngx.exit(400)                                                     
    end
    if string.find(post['referrer_domain'], " ") or string.find(post['referrer_domain'], ";") or string.find(post['referrer_domain'], "UNION") or string.find(post['referrer_domain'], "/") then
        ngx.header['x-reason'] = 'naughtyreferrer'
        ngx.exit(400)                                                     
    end    

    -- URLs
    if post['referrer'] ~= "" and not string.match(post['referrer'], '[a-z]*://[^ >,;]*') then
       ngx.header['x-reason'] = 'nourl'
       ngx.exit(400)
    end
    
    -- Session ID
    if tableHasKey(post, "sess_id") and post["sess_id"] ~= "none" and not string.match(post["sess_id"], '[%a%d][%a%d][%a%d][%a%d]%-[%a%d][%a%d][%a%d][%a%d]') then
        ngx.header['x-reason'] = "sess"
        ngx.exit(400)
    end    
    
    -- State
    local states = {"PageView", "ready", "playing"}
    if tableHasKey(post, "state") and not table_contains(states, post["state"]) then
        ngx.header['x-reason'] = "state"
        ngx.exit(400)
    end        
    
    
    -- Numbers
    if not tonumber(post['timezone']) then
       ngx.header['x-reason'] = 'tz'
       ngx.exit(400)
    end
    if post['timezone'] > 1440 or post['timezone'] < -1440 then
       ngx.header['x-reason'] = 'tzrange'
       ngx.exit(400) 
    end
    
    if not tonumber(post['responseTime']) then
       ngx.header['x-reason'] = 'resp'
       ngx.exit(400)
    end    
    if post['responseTime'] > 900000 or post['responseTime'] < 0 then
       ngx.header['x-reason'] = 'resprange'
       ngx.exit(400) 
    end
end


local server = ngx.var.influx_ip
local port = ngx.var.influx_port
local measurement = ngx.var.measurement
local db = ngx.var.db
local mode = ngx.var.mode

if mode == "write_search" then
    local lp = build_searchterm_lp(ngx, measurement)
    write_lp(server, port, lp, db)
    ngx.say(lp)
else
    if mode == "write_enhanced" then
        local lp = build_enhanced_lp(ngx, measurement)
        write_lp(server, port, lp, db)
        ngx.say('')        
    else
        if mode == "write_404" then
            local lp = build_404_lp(ngx, measurement)
            write_lp(server, port, lp, db)
            ngx.say('')                    
        else
            local lp, result = build_lp(ngx, measurement)
            write_lp(server, port, lp, db)
            
            if not result then
            ngx.status = 203
            ngx.header['Access-Control-Allow-Origin'] = '*'
            ngx.header['Access-Control-Allow-Methods'] = 'POST, OPTIONS'
            ngx.say("unoh")
            ngx.exit(203)
            end
            
            ngx.say(lp)
        end
    end
end
